#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011, 2012, 2013 OUI Technology Ltd.
# Copyright (C) 2019-2021, 2024 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Update Wiking CMS attachment thumbnails in the database.

Thumbnails are normally updated automatically on source image changes, but when
the administrator decides to change the thumbnail sizes (configuration option
image_thumbnail_sizes of Wiking CMS) or resized image size (configuration
option image_screen_size), thumbnails and rezized images must be regenerated to
match the new settings.

"""
import sys
import getopt
import os
import io
import PIL.Image

import pytis
import pytis.util
import pytis.data as pd
import wiking
import wiking.cms


def usage(msg=None):
    sys.stderr.write("""Update Wiking attachment thumbnails in the database.
Usage: %s [options]
Options: Pytis command line options, such as --config or --dbhost and --dbname.
""" % sys.argv[0])
    if msg:
        sys.stderr.write(msg)
        sys.stderr.write('\n')
    sys.exit(1)


def resize(image, size):
    img = image.copy()
    img.thumbnail(size, PIL.Image.LANCZOS)
    stream = io.BytesIO()
    img.save(stream, image.format)
    return stream.getvalue(), img.size


def run():
    if '--help' in sys.argv:
        usage()
    try:
        pytis.config.add_command_line_options(sys.argv)
        if len(sys.argv) > 1:
            usage()
    except getopt.GetoptError as e:
        usage(e.msg)
    wiking.cfg.user_config_file = pytis.config.config_file
    pytis.config.dblisten = False
    pytis.config.log_exclude = [pytis.util.ACTION, pytis.util.EVENT,
                                pytis.util.DEBUG, pytis.util.OPERATIONAL]
    while True:
        try:
            data = pd.dbtable('cms_page_attachments',
                              ('attachment_id', 'filename', 'width', 'height',
                               'image', 'image_width', 'image_height', 'thumbnail',
                               'thumbnail_size', 'thumbnail_width', 'thumbnail_height'),
                              pytis.config.dbconnection)
        except pd.DBLoginException as e:
            if pytis.config.dbconnection.password() is None:
                import getpass
                login = pytis.config.dbuser
                password = getpass.getpass("Enter database password for %s: " % login)
                pytis.config.dbconnection.update_login_data(user=login, password=password)
        else:
            break
    image_screen_size = wiking.cms.cfg.image_screen_size
    image_thumbnail_sizes = wiking.cms.cfg.image_thumbnail_sizes
    transaction = pd.transaction()
    data.select(transaction=transaction)
    try:
        while True:
            row = data.fetchone()
            if row is None:
                break
            ext = os.path.splitext(row['filename'].value())[1].lower()
            path = os.path.join(wiking.cms.cfg.storage, pytis.config.dbname, 'attachments',
                                row['attachment_id'].export() + (ext or '.'))
            attachment = open(path, 'rb')
            try:
                image = PIL.Image.open(attachment)
            except IOError as e:
                continue
            sys.stderr.write("Resizing %s (%dx%d): " %
                             (row['filename'].value(), image.size[0], image.size[1]))
            thumbnail_size = row['thumbnail_size'].value()
            if thumbnail_size is None:
                thumbnail_value, real_thumbnail_size = None, (None, None)
            else:
                if thumbnail_size == 'small':
                    size = image_thumbnail_sizes[0]
                elif thumbnail_size == 'medium':
                    size = image_thumbnail_sizes[1]
                else:
                    size = image_thumbnail_sizes[2]
                thumbnail_value, real_thumbnail_size = resize(image, (size, size))
                sys.stderr.write("%dx%d, " % real_thumbnail_size)
            resized_image_value, resized_image_size = resize(image, image_screen_size)
            sys.stderr.write("%dx%d\n" % resized_image_size)
            values = dict(
                width=image.size[0],
                height=image.size[1],
                thumbnail=thumbnail_value,
                thumbnail_width=real_thumbnail_size[0],
                thumbnail_height=real_thumbnail_size[1],
                image=resized_image_value,
                image_width=resized_image_size[0],
                image_height=resized_image_size[1],
            )
            r = pd.Row([(key, pd.Value(row[key].type(), value)) for key, value in values.items()])
            data.update(row['attachment_id'], r, transaction=transaction)
    except Exception:
        try:
            transaction.rollback()
        except Exception:
            pass
        sys.stderr.write("Transaction rolled back.\n")
        raise
    else:
        sys.stderr.write("Transaction commited.\n")
        transaction.commit()
    transaction.close()


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2018 OUI Technology Ltd.
# Copyright (C) 2019-2020 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Wiking Content Management System application definition.

The CMS application is defined as an implementation of Wiking Application Interface.

"""
import lcg
import wiking
from wiking.cms import CMSExtension, CMSExtensionModule, Roles, Users, text2content

_ = lcg.TranslatableTextFactory('wiking-cms')


class AdminControl(wiking.TopBarControl):

    def _icon(self, context):
        if wiking.module.Application.preview_mode(context.req()):
            return 'gear-larger-red'
        else:
            return 'gear-larger'

    def _menu_title(self, context):
        if wiking.module.WikingManagementInterface.authorized(context.req()):
            return _("Website Administration")
        else:
            return _("Content Management")

    def _menu_items(self, context):
        req = context.req()
        items = []
        if wiking.module.WikingManagementInterface.authorized(req):
            if not req.wmi:
                items.append(lcg.PopupMenuItem(_("Enter the Management Interface"), uri='/_wmi/',
                                               icon='gear-icon'))
            else:
                items.append(lcg.PopupMenuItem(_("Leave the Management Interface"), uri='/',
                                               icon='circle-out-icon'))
        if wiking.module.Application.preview_mode_possible(req):
            # Translators: There are two modes of operation in the CMS
            # management.  The "Production Mode" displays only the content
            # publically visible to the website visitors.  "Preview Mode",
            # on the other hand, displays also the content which is not
            # published yet.  This content is only visible to the
            # administrators until it is published.  Please make sure you
            # translate these two modes consistently acros all their
            # occurences.
            if wiking.module.Application.preview_mode(req):
                label, value = _("Switch to Production Mode"), '0'
            else:
                label, value = _("Switch to Preview Mode"), '1'
            param = wiking.module.Application._PREVIEW_MODE_PARAM
            items.append(lcg.PopupMenuItem(label, uri='%s?%s=%s' % (req.uri(), param, value),
                                           icon='refresh-icon'))
        if not req.wmi:
            if hasattr(req, 'page_write_access') and req.page_write_access:
                items.append(lcg.PopupMenuItem(_("Edit the Current Page"),
                                               uri=req.uri() + '?action=update',
                                               icon='edit-icon'))
            if req.check_roles(wiking.cms.Roles.CONTENT_ADMIN):
                items.append(lcg.PopupMenuItem(_("Create a New Page"), uri='/?action=insert',
                                               icon='create-icon'))
        if not items and wiking.cms.cfg.always_show_admin_control:
            items.append(lcg.PopupMenuItem(_("Log in for site administration"),
                                           uri='/?command=login', icon='circle-in-icon'))
        return items


class Application(wiking.Application):

    _MAPPING = dict(
        wiking.Application._MAPPING,
        _wmi='WikingManagementInterface',
        _registration='Registration',
        _sitemap='SiteMap',
    )

    _PREVIEW_MODE_COOKIE = 'wiking_cms_preview_mode'
    _PREVIEW_MODE_PARAM = '_wiking_cms_preview_mode'

    def preview_mode(self, req):
        """Query the current state of preview mode.

        Returns True if the application is currently in a
        preview mode and False when in production mode (the oposite of preview
        mode).

        These modes determine which content is displayed.  Preview mode is for
        site administrators (also unpublished content is displayed), production
        mode is for visitors (only published content is displayed).  Switching
        may be used by adimnistrators to check the differences.

        """
        if self.preview_mode_possible(req):
            return req.cookie(self._PREVIEW_MODE_COOKIE) == '1'
        else:
            return False

    def preview_mode_possible(self, req):
        """Return true if the current user is allowed to switch to the preview mode.

        Doesn't depend on the current mode, just indicates the possibility.

        """
        return req.check_roles(Roles.CONTENT_ADMIN) or req.__dict__.get('page_write_access', False)

    def set_preview_mode(self, req, value):
        """Change the current state of preview mode.

        The current mode is changed according to given 'value' -- switched to
        preview mode when 'value' is True or switched to production mode when
        value is False.

        """
        if self.preview_mode(req) != value:
            if value:
                message = _("Switching to preview mode")
            else:
                message = _("Switching to production mode")
            req.message(message, req.WARNING)
            req.set_cookie(self._PREVIEW_MODE_COOKIE, value and '1' or '0')

    def _resolve_modname(self, req):
        """Return the wiking module name to handle this request.

        Determine the name of the module responsible for handling this request.
        The module is most typically determined from the request path and/or
        request parameters.  The relavant part of 'req.unresolved_path' must be
        consumed.

        This method may be overriden in derived classes.

        """
        if req.unresolved_path:
            try:
                modname = self._mapping[req.unresolved_path[0]]
            except KeyError:
                return 'Pages'
            # Consume the unresolved path if it was in static mapping or
            # leave it for further resolution when passing to Pages.
            del req.unresolved_path[0]
            return modname
        elif req.param('action'):
            if req.param('_manage_cms_panels') == '1':
                return 'Panels'
            else:
                return 'Pages'
        else:
            return None

    def handle(self, req):
        req.wmi = False  # Will be set to True by `WikingManagementInterface' if needed.
        preview_mode_param = req.param(self._PREVIEW_MODE_PARAM)
        if preview_mode_param is not None:
            req.set_cookie(self._PREVIEW_MODE_COOKIE, preview_mode_param == '1' and '1' or None)
        wiking.module.CachedTables.reload_info(req)
        wiking.module.Config.configure(req)
        modname = self._resolve_modname(req)
        if modname:
            return req.forward(wiking.module(modname))
        else:
            try:
                return super(Application, self).handle(req)
            except wiking.Forbidden:
                # The parent method raises Forbidden when there are no menu items to redirect to.
                if req.check_roles(Roles.CONTENT_ADMIN):
                    # Give the administrator some hints on a fresh install.
                    if wiking.module.Pages.empty(req):
                        return wiking.Document(
                            title=_("Welcome to Wiking CMS"),
                            content=lcg.Container((lcg.p("There are currently no pages."),
                                                   lcg.p(lcg.link("/?action=insert",
                                                                  _("Create a new page"))))),
                        )
                    elif not self.preview_mode(req):
                        req.message(_("There are no published pages. "
                                      "You need to switch to the Preview mode "
                                      "to be able to access the unpublished pages."),
                                    req.WARNING)
                raise

    def module_uri(self, req, modname):
        """Return the base URI of given Wiking module (relative to server root).

        This method implements the interface defined by
        'wiking.Application.module_uri()' specifically for the Wiking CMS
        application.


        The method bahaves as follows:

          1. Static URI mapping (see 'wiking.Application._MAPPING') is searched
             first.  If the module is found there, the corresponding path is
             returned.
          2. If the above fails, the module is searched within CMS pages as
             their extension module.  If the module is found as an extension
             module of a particular page, the path to that page (including the
             subpath to the module) is returned.  Beware that if the same
             module had been used as an extension module for more than one
             page, there would be no way to distinguish which page to use to
             form the path and thus None is returned in such cases.
          3. If the above fails and the module is derived from
             'CMSExtensionModule', its parent module is searched according to
             2. and if found, the corresponding path plus the path to the
             submodule is returned.
          4. If the above fails and the module is accessible through the Wikimg
             Management Interface, the WMI uri is returned.
          5. If all the above fails, None is returned.  Particularly, this
             happens for modules, which are not directly associated with any
             page, which may also be the case for modules accessible through
             bindings to other modules.

         The mapping used in step 1. is called static, because it is a
         hardcoded assignment of URIs of modules needed for Wiking CMS to run
         (such as 'Documentation', 'Resources', etc).  The user is not able to
         change this mapping.  The convention is, that URIs in the static
         mapping in Wiking CMS start with an underscore to prevent conflicts
         with user defined URIs (identifiers) of CMS pages (which are dynamic
         from this perspective — the user may change them).

         Examples (calling through 'wiking.Request.module_uri()'):

           req.module_uri('Documentation')

         Returns '/_doc'.

           req.module_uri('Users')

         Returns '/users' if the module 'Users' is used in a CMS page with an
         identifier 'users'.  If the module is not used in any CMS page
         '/_wmi/users/Users' is returned.

           req.module_uri('Planner')

         Returns '/events' if the module 'Planner' is used in a CMS page with an
         identifier 'events' or None if it is not used.

           req.module_uri('BugComments')

         Returns '/bts/bug-comments' if the CMS extension module
         'BugTrackingSystem' is used in a page with an identifier 'bts' and
         'BugComments' is a submodule of 'BugTrackingSystem' with a static
         subpath 'bug-comments').

        """
        # Try the static mapping first.
        uri = super(Application, self).module_uri(req, modname)
        if uri is None:
            # Try if the module is directly embedded in a page.
            uri = wiking.module.Pages.module_uri(req, modname)
            if uri is None:
                # If not embeded directly, try if it is a submodule of an embedded module.
                mod = wiking.module(modname)
                if isinstance(mod, CMSExtensionModule):
                    parent = mod.parent()
                    if parent is None:
                        # Hack: Instantiate all CMSExtension modules to get
                        # the parent, as parentship is initialized on
                        # CMSExtension module for all its child
                        # CMSExtensionModule submodules.
                        for modname_, modcls in wiking.cfg.resolver.walk(CMSExtension):
                            wiking.module(modname_)
                        parent = mod.parent()  # It's hopefully not None now...
                    if parent is not None:
                        uri = parent.submodule_uri(req, modname)
                if uri is None:
                    uri = wiking.module.WikingManagementInterface.module_uri(req, modname)
        return uri

    def site_subtitle(self, req):
        if req.wmi:
            return _("Management Interface")
        else:
            return wiking.cfg.site_subtitle

    def menu(self, req):
        modname = req.wmi and 'WikingManagementInterface' or 'Pages'
        return wiking.module(modname).menu(req)

    def panels(self, req, lang):
        return wiking.module.Panels.panels(req, lang)

    def languages(self):
        return wiking.module.Languages.languages()

    def stylesheets(self, req):
        return wiking.module.StyleSheets.stylesheets(req)

    def body_class_names(self, req):
        if self.preview_mode(req):
            return ('preview-mode',)
        else:
            return ('production-mode',)

    def authenticate(self, req, login, password, auth_type):
        user = wiking.module.Users.user(req, login)
        if user is None and wiking.cms.cfg.allow_registration:
            # It is possible, that the user doesn't exist in the
            # application specific users table, but exists in the base
            # table of wiking CMS (the user was registered for some other
            # application sharing the same database).  Here we test if
            # that's the case and handle the situation in the login_hook()
            # below.
            u = wiking.module('wiking.cms.Users').user(req, login)
            if u and u.state() != Users.AccountState.NEW:
                user = u
        if user:
            stored_password = user.data()['password'].value()
            if wiking.cms.cfg.password_storage.check_password(password, stored_password):
                return user
        wiking.module.LoginFailures.failure(req, login, auth_type)
        return None

    def login_hook(self, req, user):
        import wiking.cms.texts
        if not wiking.module.Users.user(req, user.login()):
            # This account needs re-registration.  See 'user()' for comments.
            regcode = wiking.module('wiking.cms.Users').regenerate_registration_code(user)
            req.message(_("User %s is already registered for another site. "
                          "Please, confirm the account for this site.", user.login()))
            raise wiking.Redirect(req.module_uri('Registration'),
                                  action='reinsert', login=user.login(), regcode=regcode)
        # This is done in the login hook to display the message only once after logging in...
        state = user.state()
        text = wiking.module.Texts.text
        if state == Users.AccountState.DISABLED:
            message = text(wiking.cms.texts.disabled)
        elif state == Users.AccountState.NEW:
            uri = req.make_uri(req.module_uri('Registration'), action='confirm', uid=user.uid())
            message = text(wiking.cms.texts.unconfirmed).interpolate(lambda x: dict(uri=uri)[x])
        elif state == Users.AccountState.UNAPPROVED:
            message = text(wiking.cms.texts.unapproved)
        else:
            return
        req.message(message, req.WARNING, formatted=True)

    def logout_hook(self, req, user):
        super(Application, self).logout_hook(req, user)
        if user is not None:
            wiking.module.CryptoKeys.clear_crypto_passwords(req, user)

    def contained_roles(self, role):
        role_sets = wiking.module.RoleSets
        if isinstance(role, (list, tuple, set)):  # role is actually role ids
            role_ids = role_sets.included_role_ids_by_role_ids(role, instances=True)
        else:
            role_ids = role_sets.included_role_ids(role, instances=True)
        return role_ids

    def registration_uri(self, req):
        if wiking.cms.cfg.allow_registration:
            return req.make_uri(req.module_uri('Registration'), action='insert')
        return None

    def password_change_uri(self, req):
        return req.make_uri(req.module_uri('Registration'), action='passwd')

    def forgotten_password_uri(self, req):
        return req.make_uri(req.module_uri('Registration'), action='reset_password')

    def login_is_email(self, req):
        return wiking.cms.cfg.login_is_email

    def right_panels_bottom_content(self, req):
        if req.check_roles(Roles.CONTENT_ADMIN):
            def content(context, element):
                g = context.generator()
                return g.form((g.button(g.span('', cls='icon plus-icon') +
                                        g.span(_("New Panel"), cls='label'), type='submit'),
                               g.hidden('action', 'insert'),
                               g.hidden('_manage_cms_panels', '1')),
                              action='/', method='GET')
            return lcg.HtmlContent(content)
        else:
            return None

    def _text_content(self, req, text):
        # Default to English to avoid raising NotAcceptable where it is not handled.
        lang = req.preferred_language(raise_error=False) or 'en'
        return wiking.module.Texts.localized_text(req, text, lang=lang)

    def top_content(self, req):
        content = super(Application, self).top_content(req)
        top_text = self._text_content(req, wiking.cms.texts.top)
        if top_text:
            content.append(lcg.Container(text2content(req, top_text), name='top-text'))
        return content

    def top_controls(self, req):
        return [
            wiking.LoginControl(),
            AdminControl(),
            wiking.LanguageSelection(),
            wiking.MaximizedModeControl(),
        ]

    def bottom_bar_left_content(self, req):
        return lcg.Link('/_sitemap', wiking.module.SiteMap.title())

    def footer_content(self, req):
        text = self._text_content(req, wiking.cms.texts.footer)
        return text2content(req, text.replace('$webmaster_address', wiking.cfg.webmaster_address))

    def login_dialog_top_content(self, req):
        return text2content(req, self._text_content(req, wiking.cms.texts.login_dialog_top_text))

    def login_dialog_bottom_content(self, req):
        return text2content(req, self._text_content(req, wiking.cms.texts.login_dialog_bottom_text))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2016 OUI Technology Ltd.
# Copyright (C) 2019-2021 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Definition of Wiking configuration options.

The class 'Configuration' defines options common for all Wiking applications.
Application specific options may be defined for particular Wiking applications
by deriving from the class 'ApplicationConfiguration' defined below.

Each option has a short description and documentation.  The convention is to
make the description as short as possible (well, as long as it is meaningful)
and put all additional information into the documentation.  The description may
be used as field label within Wiking CMS web based configuration.  The
documentation is used as field description in this case.  The options which are
available through Wiking CMS web based configuration have the description and
documentation marked for translation (using the underscore function).  Another
usage of these values in the automatically generated configuration, where the
description is used as a section title (and very long titles will look odd as
well).

The docstring of the 'Configuration' class below is also used within the
documentation.  That's also why this docstring as well as option documentation
strings may use LCG formatting (the documentation is generated by LCG).

"""

import os

import lcg
import pytis.util
from wiking import Exporter, Theme

_ = lcg.TranslatableTextFactory('wiking')

pc = pytis.util.Configuration


class Configuration(pytis.util.Configuration):
    """Wiking Configuration.

    Wiking configuration options can be set through:
      * global configuration file
      * site specific configuration files
      * environment variables
      * Wiking Management Interface in Wiking CMS (when used)

    The configuration options defined in the later named sources override the
    values of any options from the previous sources.

    The global configuration may be practical for customizing global defaults.
    The following locations are searched:

      * /etc/wiking/config.py
      * /etc/wiking.py
      * /usr/local/etc/wiking/config.py
      * /usr/local/etc/wiking.py

    First of the named files which exists is used.

    Site specific configuration file is set through the environment variable
    =wiking.config_file=.  With uWSGI this is done using the following option
    in the uWSGI INI file:

    -----
    route-run = addvar:wiking.config_file=/var/www/yoursite/config.py
    -----

    Configuration files use Python syntax to assign values to configuration
    options.  The supported options are described below.  Options, which are
    not set through either method will retain their default value (listed for
    each option).  It is usually necessery to reload the application after
    configuration file changes.

    Setting options through environment variables may be useful if you prefer
    controlling them from web server's configuration files, uWSGI INI files or
    any other place where you are able to set the environment for your
    processes.  Such options take precedence over the options set through the
    configuration files.  The name of the environment variable consists of the
    prefix "=wiking.=" and the option name (all in lower case).  This approach
    has some limitations and special rules, because environment variables are
    just strings, not Python:

      * You can not pass Python objects — only primitive types, such as
        strings, numbers, boolean values and their sequences are supported.
      * The allowed values for boolean options are 'Yes', 'No', 'True',
        'False', 'On', 'Off' and are case insensitive.
      * String values must be enclosed in double quotes if they contain a
        space.
      * Sequence values may use a comma or colon to separate the items. Spaces
        between items are allowed, but remember to put quotes around.
      * =translation_path= is /appended/ to the path defined by configuration
        files.
      * =resource_path= is /added in front/ of the path defined by
        configuration files.

    Examples:
    -----
      # Apache's configuration file (inside the <VirtualHost> directive):
      SetEnv wiking.dbhost 10.0.0.122
      SetEnv wiking.dbport 5433
      # NGINX configuration file (inside the server {} block):
      uwsgi_param wiking.maintenance True;
      uwsgi_param wiking.modules 'biblio, wiking.cms'
      # uWSGI INI file:
      route-run = addvar:wiking.translation_path=/some/path:/other/path
      route-run = addvar:wiking.resource_path=/var/www/myapp/resources
    -----

    """

    class _Option_config_file(pc.StringOption, pc.HiddenOption):
        _DESCR = "Wiking global configuration file location"

        def default(self):
            for filename in ('/etc/wiking/config.py', '/etc/wiking.py',
                             '/usr/local/etc/wiking/config.py', '/usr/local/etc/wiking.py'):
                if os.access(filename, os.F_OK):
                    return filename
            return None

    class _Option_user_config_file(pc.StringOption, pc.HiddenOption):
        _DESCR = "Site specific configuration file location"

        def default(self):
            try:
                import wikingconfig
            except ImportError:
                return None
            filename = wikingconfig.__file__
            if filename.endswith('.pyc') or filename.endswith('.pyo'):
                filename = filename[:-1]
            return filename

    class _Option_modules(pc.Option):
        _DESCR = "Wiking module search order"
        _DOC = ("A sequence of names of Python modules (as strings) used to search for Wiking "
                "module class definitions.  The named Python modules will be searched in given "
                "order and they must be available through Python path.")
        _DEFAULT = ('wiking.cms', )

    class _Option_dbname(pc.StringOption):
        _DESCR = "Database name"
        _DOC = ("Name of the database to connect to.  If not defined, the server name of the "
                "current virtual host will be used.")

    class _Option_dbhost(pc.StringOption):
        _DESCR = "Database host"
        _DOC = ("If None (default) Wiking will connect to the database through UNIX domain "
                "sockets.  If a hostname or IP address is specified, Wiking will connect "
                "to given host through a network socket.")

    class _Option_dbport(pc.NumericOption):
        _DESCR = "Database port number"

    class _Option_dbuser(pc.StringOption):
        _DESCR = "Database user"

    class _Option_dbpass(pc.StringOption):
        _DESCR = "Database password"

    class _Option_dbsslm(pc.StringOption):
        _DESCR = "Database ssl mode"

    class _Option_dbschemas(pc.StringOption):
        _DESCR = "List of database schemas to use in the order of their preference"

    class _Option_dbconnections(pc.Option):
        _DESCR = "Alternative database connections."
        _DOC = ("Wiking modules normally use the default database connection defined by 'dbname', "
                "'dbhost', 'dbport', 'dbuser', 'dbpass' and 'dbsslm', "
                "but certain applications may require "
                "multiple database connections.  If your installation uses modules employing "
                "alternative connections, you must configure them using this option.  The value is "
                "a dictionary assigning a connection specification to each connection by name. "
                "Connection names are defined by applications (each application should mention the "
                "names of used conections in its documentation).  The connection specification "
                "(the value assigned to a connection name) is a dictionary with keys 'dbname', "
                "'dbhost', 'dbport', 'dbuser' and 'dbpass'.  Only 'dbname' is mandatory.  Their "
                "meaning and default values are the same as for the configuration options of the "
                "same names specifying the properties of the default connection.")
        _DEFAULT = {}

    class _Option_maintenance(pc.BooleanOption):
        _DESCR = "Maintenance mode flag"
        _DOC = ("Setting this value to True will tell Wiking to run in the maintenance mode. "
                "This mode is mostly useful for server administration tasks, which require "
                "exclusive access to the database.  Wiking will not attempt to connect to "
                "the database in this mode and will display a polite error message in response "
                "to all requests.")
        _DEFAULT = False

    class _Option_smtp_server(pc.StringOption):
        _DESCR = "Name or address of SMTP server"
        _DOC = ("SMTP server is used for sending bug reports by e-mail.  To allow this feature, "
                "you must also supply the 'bug_report_address' option.")
        _DEFAULT = 'localhost'

    class _Option_smtp_port(pc.NumericOption):
        _DESCR = "Port for SMTP server"
        _DOC = ("SMTP server port")
        _DEFAULT = 25

    class _Option_allow_smtp_email_validation(pc.BooleanOption):
        _DESCR = "Allow SMTP e-mail validation"
        _DOC = ("Wiking email validation functions may use SMTP to verify the existence of an "
                "e-mail account on the remote mail server.  Some installations, however, may "
                "not allow the server to create such SMTP connections (e.g. when the server "
                "is located behind a firewall) so this option is off by default.  If not "
                "enabled, the mail addres domain is only checked through DNS.")
        _DEFAULT = False

    class _Option_debug(pc.BooleanOption):
        _DESCR = "Debugging mode"
        _DOC = ("Turn on the debugging mode to allow sending detailed exception tracebacks to "
                "the browser rather than logging it to servers error log.  Note, that tracebacks "
                "may contain sensitive information, so this mode should never be on in a "
                "production environment.  It is only intended for developers.")
        _DEFAULT = False

    class _Option_debug_transactions(pc.BooleanOption):
        _DESCR = "Debug pending transactions"
        _DOC = ("Turn on logging tracebacks of pending transactions after each request. "
                "This option should be used for development only.  Application in production "
                "environment should have no pending transactions when a request is finished.")
        _DEFAULT = False

    class _Option_profile(pc.BooleanOption):
        _DESCR = "Profiling mode"
        _DOC = ("Turn on the profiling mode to log profiling information. "
                "It only works when debugging mode is enabled as well. "
                "This option is intended only for developers.")
        _DEFAULT = False

    class _Option_log_format(pc.StringOption):
        _DESCR = "Logging format"
        _DOC = ("""Python format string used for printing error message to the system log.

                The format string may use the following format variables:
                  error_type -- error class name string (the class derived from 'RequestError'),
                  server_hostname -- requested server host name (virtual host),
                  uri -- URI of the request,
                  abs_uri -- Absolute URI of the request (including server URI),
                  method -- HTTP method of the request,
                  user -- current user's login name ('User.login()') or 'anonymous' when
                    user is not logged,
                  remote_host -- IP adress of the client's host as a string,
                  referer -- HTTP referer (URI of the page linking to this request's
                    URI),
                  user_agent -- client software identification from the 'User-Agent' HTTP
                    header,
                  server_software -- server software identification (current versions of
                    Wiking, LCG and Pytis)
                  ref_id -- Reference identifier for pairing the log entry with bug report
                    email (see 'Application.report_error()').

                """)
        _DEFAULT = "%(error_type)s[%(ref_id)s]: %(method)s %(abs_uri)s [%(user)s@%(remote_host)s]"

    class _Option_bug_report_address(pc.StringOption):
        _DESCR = "E-mail address where bug reports will be sent"
        _DOC = ("Tracebacks of uncaught exceptions are sent automatically by e-mail to the "
                "specified address.  If None, bug reports are not sent by e-mail.  In any case "
                "they are logged to server's error log.")
        _DEFAULT = None

    class _Option_webmaster_address(pc.StringOption):
        _DESCR = _("Webmaster's e-mail address")
        _DOC = ("Webmaster's address is used as public contact address for the site. It is "
                "displayed at the bottom of each page, in error messages, RSS feeds and so on. "
                "If no address is given, it will be automatically set to 'webmaster@DOMAIN', "
                "where DOMAIN is the domain part of the server hostname if the hostname starts "
                "with 'www.' or the whole server hostname in other cases.  Please make sure "
                "that this address is valid (e-mail sent to it is delivered to a responsible "
                "person).")
        _DEFAULT = None

    class _Option_default_sender_address(pc.StringOption):
        _DESCR = _("Default sender address")
        _DOC = ("E-mail messages sent by the system, such as automatic notifications, password "
                "reminders, bug-reports etc. will use this sender address.  If no address is "
                "given, it will be automatically set to 'wiking@DOMAIN', where DOMAIN is the "
                "domain part of the server hostname if the hostname starts with 'www.' or the "
                "whole server hostname in other cases.  Please make sure that this address is "
                "valid (e-mail sent to it is delivered to a responsible person).")
        _DEFAULT = None

    class _Option_https_ports(pc.Option):
        _DESCR = "HTTPS port numbers"
        _DOC = ("The default HTTPS port is 443 but certain server configurations may require "
                "using a different port or several ports for HTTPS connections.")
        _DEFAULT = (443,)

    class _Option_force_https_login(pc.BooleanOption):
        # Translators: ``Force'' means ,,do not allow otherwise'',
        # sentence in imperative followed by a checkbox.
        _DESCR = _("Force HTTPS login")
        _DOC = _("If enabled, the login form will always be redirected to an HTTPS address "
                 "to ensure the security of the submitted credentials.  This, however, requires "
                 "your server to be setup to accept HTTPS requests for the same virtual host and "
                 "pass them to Wiking.")
        _DEFAULT = False

    class _Option_allow_http_authentication(pc.BooleanOption):
        _DESCR = _("Allow HTTP authentication")
        _DOC = ("Deprecated.  Define 'authentication_providers' instead.")
        _DEFAULT = True

    class _Option_authentication_providers(pc.BooleanOption):
        _DESCR = _("Allow HTTP authentication")
        _DOC = ("List of 'wiking.AuthenticationProvider' instances to use for user "
                "authentication in given order.  Wiking by default supports cookie "
                "based authentication through 'wiking.CookieAuthenticationProvider'"
                "and HTTP Basic authentication through "
                "'wiking.HTTPBasicAuthenticationProvider'.  Other authentication "
                "schemes may be implemented and added to this list.")
        def default(self):
            import wiking
            providers = [wiking.CookieAuthenticationProvider()]
            if self._configuration.allow_http_authentication:
                providers.append(wiking.HTTPBasicAuthenticationProvider())
            return providers

    class _Option_translation_path(pc.Option):
        _DESCR = "Translation search path"
        _DOC = ("The value is a sequence of directory names (strings), where locale data are "
                "searched.  These directories depend on your installation.  Each directory "
                "should contain a subdirectory 'lang/LC_MESSAGES' and a file 'domain.mo' in it, "
                "where lang is the language code and domain is the translation domain name.")

        def default(self):
            return ('/usr/local/share/wiking/translations',
                    '/usr/local/share/lcg/translations',
                    '/usr/local/share/pytis/translations')

    class _Option_resource_path(pc.Option):
        _DESCR = "Resource search path"
        _DOC = ("The value is a sequence of directory names (strings), where resource files, "
                "such as images, media or scripts are searched.  The directories are searched "
                "in given order.  Files may be put into type specific subdirectories (jpegs and "
                "pngs into 'images', stylesheets into 'css', javascripts into 'scripts') or to "
                "the resource directories directly.  Wiking will return a 404 error and log the "
                "exact list of directories where a particular file is searched when it is unable "
                "to locate it.  Beware that all files located within the named directories will "
                "be directly exposed to the Internet!")

        def default(self):
            return ('/usr/local/share/wiking/resources',
                    '/usr/local/share/lcg/resources',
                    '/usr/local/share/pytis/resources')

    class _Option_doc_dirs(pc.Option):
        _DESCR = "Documentation search directories"
        _DOC = ("The value is a dictionary of directory names keyed by component "
                "identifiers.  Component identifiers are unique strings identifying different "
                "components of a Wiking application.  Core Wiking requires directories for two "
                "components: 'wiking' and 'lcg'.  Wiking applications may require access to "
                "documentation of additional components (application specific documentation). "
                "Dictionary values are names of directories containing documentation source "
                "files, such as full path to the 'doc/src' subdirectory of the wiking package.")

        def default(self):
            return {'wiking': '/usr/local/share/wiking/doc/src',
                    'lcg': '/usr/local/share/lcg/doc/src'}

    class _Option_server_hostname(pc.StringOption):
        _DESCR = "Primary server host name"
        _DOC = ("This value may be used by applications to distinguish "
                "between separate servers.  One server may still have multiple "
                "aliases so it can be accessed under different server names (as "
                "returned by 'wiking.Request.server_hostname()'), but this "
                "configuration option will have the same value for all servers "
                "which belong to one primary server (or more precisely, all "
                "servers, which share the same configuration).  This is usually "
                "also the primary (official) server name.  The primary server, "
                "however, is typically still a virtual server and there may be "
                "several primary servers on one physical host.")
        _DEFAULT = None

    class _Option_site_title(pc.StringOption):
        # Translators: Site means a webpage or web application.
        _DESCR = _("Site title")
        _DOC = _("Site title is a short and (hopefully) unique title of the whole website. "
                 "It will appear at the top of every page.")
        _DEFAULT = 'Wiking site'

    class _Option_site_subtitle(pc.StringOption):
        # Translators: Site means a webpage or web application.
        # Subtitle as in ,,title -- subtitle''.
        _DESCR = _("Site subtitle")
        _DOC = _("Site subtitle is an optional more descriptive title of the website.  It will "
                 "appear at the top of every page together with site title, but where brevity "
                 "matters, only title will be used.")
        _DEFAULT = None

    class _Option_site_icon(pc.StringOption):
        # Translators: Web page or web application specific image.
        _DESCR = _("Site icon")
        _DOC = ("Full file path to the site icon to be served as '/favicon.ico'. "
                "See http://en.wikipedia.org/wiki/Favicon for more information. "
                "Only ICO format icons are currently supported.")
        _DEFAULT = None

    class _Option_site_image(pc.StringOption):
        _DESCR = "Open Graph site image URL"
        _DOC = ("URL of the site image/logo to be used for Open Graph meta data. "
                "See http://en.wikipedia.org/wiki/Facebook_Platform#Open_Graph_protocol "
                "for more information.  Note, that this is a URL, not file path as for "
                "the site_icon.")
        _DEFAULT = None

    class _Option_theme(pc.Option):
        _DESCR = "Color theme"
        _DOC = ("Color theme is a set of predefined colors to be used for style sheet color "
                "substitution.  Thus Wiking style sheets may be color neutral and colors may "
                "be defined separately.  The value is an instance of 'wiking.Theme' class.")
        _DEFAULT = Theme()

    class _Option_extra_scripts(pc.StringOption):
        _DESCR = "Site specific scripts"
        _DOC = ("Sequence of filenames of additional site specific JavaScript files to be "
                "loaded with every page.  The named files must be available within resource "
                "directories as configured by the option 'resource_path'.")
        _DEFAULT = ()

    class _Option_default_language(pc.StringOption):
        _DESCR = _("Default language")
        # Translators: Language negotiation means the process where the server is trying
        # to find a language that matches with what the users internet browser asks for.
        _DOC = _("Defines the language (lowercase ISO 639-1 Alpha-2 code) to be used when "
                 "content language negotiation fails. If not set (which is the default), the "
                 "server will issue a 606 HTTP error (Not acceptable) when the document is "
                 "not available in any of the requested languages. The error page will contain "
                 "a description of the problem and a list of available language variants of "
                 "the requested document. The negotiation is based on the browser setting of "
                 "user's language preference passed through the 'Accept-Language' HTTP header. "
                 "If set, Wiking will automatically serve given language variant instead of "
                 "issuing the error if that variant is available.  If unavailable, the 606 "
                 "error is raised anyway, so you need to make sure that all documents are "
                 "available at least in the default language variant if you want to avoid 606 "
                 "errors completely.")
        _DEFAULT = None

    class _Option_default_language_by_domain(pc.Option):
        _DESCR = "Default language by server domain name"
        _DOC = ("Allows setting different default language based on the server domain name.  The "
                "value is a dictionary, where the server name is a key and language code is "
                "a value.  If the server domain name of the current request is found in the "
                "dictionary, the assigned value overrides the value of 'default_language' for "
                "given request.")
        _DEFAULT = {}

    class _Option_language_selection_image(pc.StringOption):
        _DESCR = "URI of images used in language selection menu."
        _DOC = ("If defined, the language selection links will include images (typically national "
                "flags).  The value is a string where '%s' is replaced by the language code of a "
                "particular language to get the URI of an image used for that language.  It is up "
                "to the developer/administrator to ensure that the resulting URIs are handled "
                "correctly and images of all used languages are available.")
        _DEFAULT = None

    class _Option_default_timezone(pc.Option):
        _DESCR = "Default time zone"
        _DOC = ("The time zone to use when obtaining client's time zone info fails. "
                "The value must be a 'datetime.tzinfo' instance.  In most situations "
                "'wiking.TZInfo' can be used.  See its docstring for usage and "
                "limitations.  When None, times will be displayed in UTC when "
                "client's time zone can not be determined.")
        _DEFAULT = None

    class _Option_session_expiration(pc.NumericOption):
        _DESCR = "Session expiration"
        _DOC = ("A number of hours to keep the login session alive.  The session is automatically "
                "terminated when the user has no activity for given time interval.")
        _DEFAULT = 2

    class _Option_persistent_sessions(pc.BooleanOption):
        _DESCR = "Persistent sessions"
        _DOC = ("When set to False (by default), the session will be ended when the browser "
                "is closed.  When True, the remote session overlaps browser life time, but "
                "in both cases, the option 'session_expiration' is honoured.")
        _DEFAULT = False

    class _Option_resolver(pc.Option):
        _DESCR = "Wiking module resolver"
        _DOC = ("Module resolver is used to locate available Wiking modules.  The value must be "
                "a 'WikingResolver' instance.  If not set, the default instance will be created "
                "by Wiking handler in the initialization phase, passing it the value of "
                "the configuration option 'modules' defined above.  If you supply your own "
                "instance, the configured value of 'modules' will have no effect.")

    class _Option_exporter(pc.Option):
        _DESCR = "Page exporter class"
        _DOC = ("Exporter is responsible for rendering the final page.  It gets the logical "
                "description of the page elements on input and outputs its representation in "
                "HTML/XHTML/XML or any other desired format.  You can make minor page display "
                "customizations, change the layout completely or use another output format by "
                "overriding the default exporter.  See LCG documentation for more information "
                "about the export mechanism.")
        _DEFAULT = Exporter

    class _Option_immediate_filters(pc.BooleanOption, pc.HiddenOption):
        _DESCR = ("Whether to apply filtering combobox selection immediately.")
        _DEFAULT = True

    class _Option_ignored_crypto_names(pc.Option):
        _DESCR = "Sequence of ignored crypto names"
        _DOC = ("Ignored crypto names don't raise decryption password dialog and "
                "show obfuscated data instead.")
        _DEFAULT = ()

    class _Option_resource_client_cache_max_age(pc.NumericOption):
        _DESCR = "Maximum age to allow clients to cache resources without re-requesting them."
        _DOC = ("If not None, the 'Resources' module (which is normally responsible for serving "
                "stylesheets, images, javascript and other static files) will allow the browser "
                "to consider the resource files being fresh once loaded for given number of "
                "seconds without querying the server again. This is done by setting the "
                "'Cache-Control' HTTP header for all resource file request responses to "
                "'max-age=n'.  This will reduce the number of requests needed for a page "
                "to load and thus reduce server load and improve user experience, but there "
                "is no way to make the browser to reload the resource when it is changed on "
                "server, so use with caution.  The default value is ten minutes.")
        _DEFAULT = 600

    class _Option_crawl_delay(pc.NumericOption):
        _DESCR = "Minimum delay in seconds between two successive requests from spiders."
        _DOC = ("If not None, the application will serve a 'robots.txt' file with 'Crawl-Delay' "
                "directive set to given number.  See "
                "http://en.wikipedia.org/wiki/Robots_exclusion_standard for more info.")
        _DEFAULT = None

    class _Option_special_cc_addresses(pc.Option):
        _DESCR = "Sequence of e-mail addresses to add to CC under certain circumstances."
        _DOC = ("Those addresses are added to CC of e-mails sent by Wiking where user id "
                "is declared.")
        _DEFAULT = ()

    class _Option_special_cc_exclude_roles(pc.Option):
        _DESCR = "Sequence of Roles to ignore for special CC."
        _DOC = ("Don't add special CC addresses when the declared user id is in any of the "
                "roles ('Roles' constants) given here.")
        _DEFAULT = ()

    class _Option_print_spec_dir(pc.StringOption):
        _DESCR = "Directory of output specifications."
        _DEFAULT = './output'

    class _Option_xsendfile_paths(pc.Option):
        _DESCR = "Sequence of filesystem directories enabled for X-SendFile downloads."
        _DOC = ("Each item is a string representing absolute path to a directory "
                "where server assisted file downloads using X-SendFile may be used.  "
                "The paths configured here must be also set up in server's "
                "configuration and the server must support this feature.  "
                "X-SendFile is supported by Apache with mod_xsendfile and "
                "lighttpd.  See 'xaccel_paths' for alternative feature supported "
                "by the Nginx server.")
        _DEFAULT = ()

    class _Option_xaccel_paths(pc.Option):
        _DESCR = "Mapping of filesystem directories to redirection URLs for X-Accel downloads."
        _DOC = ("Each item of the sequence is a pair of strings, where the "
                "first string is a directory path and the second string is "
                "the corresponding URI.  These pairs must match the locations "
                "configured for internal redirection in server's configuration. "
                "X-Accel redirection is supported by the Nginx server.  See "
                "'xsendfile_path' for alternative feature supported "
                "by Apache and lighttpd servers.")
        _DEFAULT = ()

    class _Option_resources_version(pc.StringOption):
        _DESCR = "String denoting version of serverd resource files."
        _DOC = ("This option makes it possible to defend against aggressive caching "
                "of resource files by some browsers or proxy servers (this sometimes "
                "happens even though the resource should be reloaded according to "
                "HTTP headers and may render the application unusable after an upgrade). "
                "When set, the resources served by a Wiking application will get "
                "a unique URI containing the given string.  For example, when not "
                "set, a CSS file has the URI '/_resources/default.css', but when "
                "'resources_version' is set to '2.0-beta', the URI will "
                "automatically change to '/_resources-2.0-beta/default.css'. "
                "So by changing the version all clients will be forced to reload "
                "all resources.  You will typically want to change the version on "
                "each deployment of a new version of an application on a "
                "production server.")
        _DEFAULT = None

    class _Option_viewport(pc.StringOption):
        _DESCR = "String value for the HTML head 'viewport' META tag."
        _DOC = ("The viewport meta tag may be used to control the page's "
                "dimensions and scaling within the browser window. Check "
                "https://developer.mozilla.org/en-US/docs/Mozilla/Mobile/Viewport_meta_tag "
                "for more information.")
        _DEFAULT = 'width=device-width, initial-scale=1'

    class _Option_show_login_control(pc.BooleanOption):
        # Translators: Yes/No configuration option label.  Should the "Log in"
        # button be displayed?
        _DESCR = _("Show login control")
        _DOC = ("The 'Log in' button, together with the login menu "
                "(registration, forgotten password, ...) is normally displayed "
                "to anonymous users (when not logged in) at the top right corner "
                "of the page.  This option may be used to disable these controls "
                "for sites where users are normally not supposed to log in or "
                "register.  This will avoid confusion and improve usability "
                "(reduce unecessary 'clutter' on the page).")
        _DEFAULT = True

    class _Option_show_submenu(pc.BooleanOption):
        _DESCR = _("Show submenu side bar")
        _DOC = ("The submenu of the current main menu item is by default displayed "
                "at the left side of every page.  This submenu allows quick access "
                "to other subsections of the current top level section, which may be "
                "practical, but it is redundant when the top menu is foldable.")
        _DEFAULT = True


class ApplicationConfiguration(pytis.util.Configuration):
    """Base class for application specific configuration.

    Wiking applications may define their own configuration options in a class
    derived from this one.  An instance of the application specific
    configuration class must be then created within the application's module
    name space.

    For example Wiking CMS defines its options in a class
    'wiking.cms.CMSConfiguration' and this class is instantiated in
    'wiking.cms.__init__py' as 'wiking.cms.cfg'.  Thus its options may be
    accessed as 'wiking.cms.cfg.password_storage' etc.

    Application specific configuration options will be automatically read from
    the same files as the other Wiking configuration options.

    """
    class _Option_config_file(pc.StringOption, pc.HiddenOption):

        def default(self):
            import wiking
            return wiking.cfg.config_file

    class _Option_user_config_file(pc.StringOption, pc.HiddenOption):

        def default(self):
            import wiking
            return wiking.cfg.user_config_file

    def __init__(self, *args, **kwargs):
        # Options are not inherited - copy them to force inheritance.
        self.__class__._Option_config_file = ApplicationConfiguration._Option_config_file
        self.__class__._Option_user_config_file = ApplicationConfiguration._Option_user_config_file
        # Delay configuration initialization until first access to avoid a dependency loop
        # when the application's Python module is imported into its config file (eg. in
        # order to use its __version__ in resources_version configuration option).  The
        # app's Pyton module creates an ApplicationConfiguration instance and this instance
        # needs to read the configuration file in time when the module import is not
        # finished yet which usually results in an AttributeError.
        self.__dict__['_delayed_init'] = lambda: pc.__init__(self, *args, **kwargs)

    def __getattr__(self, name):
        if self._delayed_init:
            self._delayed_init()
            self._delayed_init = None
        return super().__getattr__(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2012, 2013, 2015 OUI Technology Ltd.
# Copyright (C) 2019 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Perform incremental upgrades of an existing Wiking CMS database.

The script must run under a user having sufficient permissions to modify the
database table.  It runs a sequence of upgrade scripts in one transaction and
rolls back all changes when one of the scripts fails.

"""
import os
import sys
import glob
import psycopg2 as dbapi


def usage(msg=None):
    message = """Perform incremental upgrades of an existing Wiking CMS database.
Usage: %s [-p port] database directory
  database ... name of the Wiking CMS database to upgrade
  directory ... path to Wiking upgrade scripts (the `upgrade' subdirectory of a
    Wiking source archive)
""" % sys.argv[0]
    if msg:
        message += msg + '\n'
    sys.stderr.write(message)
    sys.exit(1)


def run(args):
    if '--help' in args:
        usage()
    if '-p' in args:
        i = args.index('-p')
        try:
            __, port = args.pop(i), int(args.pop(i))
        except (IndexError, ValueError):
            usage("Argument -p requires a number.")
    else:
        port = 5432
    try:
        database, directory = args[1:]
    except ValueError:
        usage("Invalid number of arguments.")
    if not os.path.isdir(directory):
        usage("Directory '%s' does not exist!" % directory)
    upgrade_scripts = sorted(glob.glob(os.path.join(directory, 'upgrade.*.sql')))
    if not upgrade_scripts:
        usage("Directory '%s' contains no upgrade scripts!" % os.path.abspath(directory))
    target_version = int(upgrade_scripts[-1].split('.')[-2])
    connection = dbapi.connect(database=database, port=port)
    try:
        cursor = connection.cursor()
        cursor.execute("select version from cms_database_version;")
        source_version = cursor.fetchone()[0]
        if source_version == target_version:
            print("The database is already at version %d.\n" % source_version)
            sys.exit(0)
        elif source_version > target_version:
            sys.stderr.write("The database is already at version %d, "
                             "but the highest upgrade script is %d.\n" %
                             (source_version, target_version))
            sys.exit(1)
        for version in range(source_version + 1, target_version + 1):
            filename = 'upgrade.%02d.sql' % version
            sql = open(os.path.join(directory, filename)).read()
            print("Applying %s ..." % filename)
            cursor.execute(sql)
        cursor.execute("update cms_database_version set version=%d;" % target_version)
        connection.commit()
    except Exception as e:
        connection.rollback()
        sys.stderr.write("Error: %s" % e)
        sys.stderr.write("Transaction rolled back.\n")
        sys.exit(1)
    else:
        print("Database %s upgraded successfully to version %d." % (database, target_version))
    finally:
        connection.close()

if __name__ == '__main__':
    run(sys.argv)



import collections
import lcg
import inspect
import re


class Reader(lcg.Reader):
    """Generate Wiking API documentation out of Python docstrings."""

    _ARG_REGEX = re.compile("^  (\w+) -- ", re.MULTILINE)

    def _title(self):
        return "Wiking Application API"

    def _variants(self):
        return ('en',)

    def _content(self, lang):
        from wiking import Application
        parser = lcg.Parser()
        # modules = [(k, v) for k, v in wiking.application.__dict__.items()
        #            if type(v) == type(Module) and issubclass(v, Module) \
        #            and v.__module__ == 'wiking.application']
        # modules.sort()
        content = parser.parse(Application.__doc__) + \
            [lcg.TableOfContents(title="Application methods:", depth=1)]
        for name, method in sorted(Application.__dict__.items()):
            if name.startswith('_') or name.startswith('action_') \
                    or not callable(method) or not method.__doc__:
                continue
            args, varargs, varkw, defaults = inspect.getargspec(method)
            title = name + inspect.formatargspec(args[1:], varargs, varkw, defaults)
            doc = self._ARG_REGEX.sub(lambda m: ":" + m.group(1) + ": ", inspect.getdoc(method))
            content.append(lcg.Section(title=title, content=parser.parse(doc), id=name))
        return dict(content=content)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Tomáš Cerha <t.cerha@gmail.com>
# Copyright (C) 2019 Tomáš Cerha <t.cerha@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Update Wiking CMS user passwords to salted PBKDF2 hashes.

May be optionally applied after upgrade of Wiking CMS database to version 72 to
convert the existing plain text or unsalted md5 passwords to salted PBKDF2 hashes.

"""
import sys
import getopt

import pytis
import pytis.util
import pytis.data as pd
import wiking
import wiking.cms


def usage(msg=None):
    sys.stderr.write("""Update Wiking CMS passwords to salted hashes.
Usage: %s [options]
Options: Pytis command line options, such as --config or --dbhost and --dbname.
""" % sys.argv[0])
    if msg:
        sys.stderr.write(msg)
        sys.stderr.write('\n')
    sys.exit(1)


def run():
    if '--help' in sys.argv:
        usage()
    try:
        pytis.config.add_command_line_options(sys.argv)
        if len(sys.argv) > 1:
            usage()
    except getopt.GetoptError as e:
        usage(e.msg)
    wiking.cfg.user_config_file = pytis.config.config_file
    wiking.cms.cfg.user_config_file = pytis.config.config_file
    pytis.config.dblisten = False
    pytis.config.log_exclude = [pytis.util.ACTION, pytis.util.EVENT,
                                pytis.util.DEBUG, pytis.util.OPERATIONAL]
    while True:
        try:
            data = pd.dbtable('users', ('uid', 'login', 'password'), pytis.config.dbconnection)
        except pd.DBLoginException as e:
            if pytis.config.dbconnection.password() is None:
                import getpass
                login = pytis.config.dbuser
                password = getpass.getpass("Enter database password for %s: " % login)
                pytis.config.dbconnection.update_login_data(user=login, password=password)
        else:
            break
    storage = wiking.Pbkdf2PasswordStorage()
    transaction = pd.transaction()
    data.select(transaction=transaction)
    n = 0
    plain = 0
    try:
        while True:
            row = data.fetchone()
            if row is None:
                break
            orig_prefix, orig_password = row['password'].value().split(':', 1)
            if orig_prefix == 'md5u':
                prefix = 'pbkdf2/md5'
            elif orig_prefix == 'plain':
                prefix = 'pbkdf2'
                plain += 1
            else:
                continue
            password = prefix + ':' + storage.stored_password(orig_password)
            data.update(row['uid'], pd.Row([('password', pd.sval(password))]),
                        transaction=transaction)
            n += 1
    except Exception:
        try:
            transaction.rollback()
        except Exception:
            pass
        sys.stderr.write("Transaction rolled back.\n")
        raise
    else:
        print("Total %d passwords updated (%d from plain text, %d from md5)." % \
            (n, plain, n - plain))
        transaction.commit()
    transaction.close()

if __name__ == '__main__':
    run()
